{.-----------------------------------------------------------------------------
  Performs operation on DCTD files
  @author l0co@wp.pl
  @version File created: 2004.10.05, 18:42:45
-----------------------------------------------------------------------------.}

unit uDataClasses;

interface

uses Classes, Contnrs, SysUtils;

const
  CNotFound = -1; //. Not found item by FindFileItem method

{------------------------------------------------------------------------------
  Types
------------------------------------------------------------------------------}

type
  TName = String[255]; //. Item name
  TFilename = String[255]; //. File name
  TTagType = (ttA, ttTitle, ttUnknown); //. Last tag type

{------------------------------------------------------------------------------
  Classes
------------------------------------------------------------------------------}

type
  TFileItem = class;

  //. Rank structure
  TRankItem = record
    Item: TFileItem;
    Rank: Integer;
  end;

  //. List of items
  TItemList = class(TObjectList)
  private
    function GetFileItems(Index: Integer): TFileItem;
    procedure SetFileItems(Index: Integer; const Value: TFileItem);
  public
    property FileItems[Index: Integer]: TFileItem read GetFileItems write SetFileItems;
  end;

  //. Single item
  TFileItem = class
  private
    FName: TName; //. Item ID (pascal name)
    FUpperName: TName; //. Item ID in upper case
    FFilename: TFilename; //. File name of item
    FOwnerCount: Integer; //. Owners count (in files structure)
  protected
    FOwner: TFileItem; //. Item's owner (in files structure)
    FInheritance: TFileItem; //. Class ancestor (in classes hierarchy)
    FContents: String; //. Original file contents
    FParsed: String; //. File contents after parsing
  public
    constructor Create(AFilename: TFilename; AOwner: TFileItem);

    function ParseWord(AWord: String): String;
    procedure Parse; virtual;
    procedure Read; virtual;
    procedure Write; virtual;

    function GetFullName: String;
    function IsMyOwner(AItem: TFileItem): Boolean;
    function IsMyInheritance(AItem: TFileItem): Boolean;

    procedure AddDebug(DebugText: String);

    property Name: TName read FName;
    property UpperName: TName read FUpperName;
    property Filename: TFilename read FFilename;
    property Owner: TFileItem read FOwner;
    property Inheritance: TFileItem read FInheritance;
    property OwnerCount: Integer read FOwnerCount;
  end;

  //. Item container (includes item fields and methods and list of descendant items)
  TFileContainer = class(TFileItem)
  private
    function IsMySubItem(Index: Integer): Boolean;
  protected
    FItemList: TItemList; //. List of descendant items
  public
    constructor Create(AFilename: TFilename; AOwner: TFileItem);
    destructor Destroy; override;

    procedure Parse; override;
    procedure Read; override;
    procedure Write; override;

    property ItemList: TItemList read FItemList;
  end;

function CompareFileItems(Item1, Item2: Pointer): Integer;
function FindFileItem(AUpperName: String): Integer;

var
  FileProject: TFileContainer; //. Main project object
  IList: TItemList; //. Sorted list of all project objects

{------------------------------------------------------------------------------
  Implementation
------------------------------------------------------------------------------}

implementation

uses uGlobals;

{.-----------------------------------------------------------------------------
  Compares two TFileItem objects alphabetically for Sort method (TListSortCompare)
-----------------------------------------------------------------------------.}
function CompareFileItems(Item1, Item2: Pointer): Integer;
begin
  if TFileItem(Item1).FUpperName = TFileItem(Item2).FUpperName then Result := 0 else
  if TFileItem(Item1).FUpperName > TFileItem(Item2).FUpperName then Result := 1 else
    Result := -1;
end;

{.-----------------------------------------------------------------------------
  Binary search of object with given UpperName
  @param AUpperName Object UpperName property
  @return Index of first element with given UpperName in IList
-----------------------------------------------------------------------------.}
function FindFileItem(AUpperName: String): Integer;

  function BinSearch(str: String; xs, xe: Integer): Integer;
  var
    mid: Integer;
  begin
    mid := (xe + xs) div 2;
    if IList.FileItems[mid].UpperName = str then begin
      Result := mid;
      Exit;
    end;
    if (xe <= xs) then begin
      Result := CNotFound;
      Exit;
    end;
    if IList.FileItems[mid].UpperName < str then
      Result := BinSearch(str, mid+1, xe) else
      Result := BinSearch(str, xs, mid-1);
  end;

begin
  Result := BinSearch(AUpperName, 0, IList.Count-1);
  if Result > 0 then
    while (Result > 0) and (IList.FileItems[Result-1].UpperName = IList.FileItems[Result].UpperName) do
      Dec(Result);
end;

{ TItemList }

{.-----------------------------------------------------------------------------
  Access method to FileItems property
-----------------------------------------------------------------------------.}
function TItemList.GetFileItems(Index: Integer): TFileItem;
begin
  Result := TFileItem(inherited Items[Index]);
end;

{.-----------------------------------------------------------------------------
  Access method to FileItems property
-----------------------------------------------------------------------------.}
procedure TItemList.SetFileItems(Index: Integer; const Value: TFileItem);
begin
  inherited Items[Index] := Value;
end;

{ TFileItem }

{.-----------------------------------------------------------------------------
  Object's constructor
  Initializing object's fields. Calculating object's OwnersCount
  @param AFilename File name to create object from
  @param AOwner Item owner (in files structure)
-----------------------------------------------------------------------------.}
constructor TFileItem.Create(AFilename: TFilename; AOwner: TFileItem);
var
  LOwner: TFileItem;
begin
  FContents := '';
  FParsed := '';
  FInheritance := nil;
  FFilename := AFilename;
  FOwner := AOwner;
  FName := GetFilenameSection(AFilename, CLastSection);
  FUpperName := UpperCase(FName);

  FOwnerCount := 0;
  LOwner := AOwner;
  while LOwner <> nil do begin
    Inc(FOwnerCount);
    LOwner := LOwner.FOwner;
  end;

  Inc(ListPosition);
  IList.Add(Self);

  // Debug log
  if AOwner = nil then
    AddDebug('create, name=' + FName + ', ownercount='+ IntToStr(OwnerCount) + ', no owner') else
    AddDebug('create, name=' + FName + ', ownercount='+ IntToStr(OwnerCount) + ', owner=' + AOwner.FName);
end;

{.-----------------------------------------------------------------------------
  Parses single word
  Searches given words in IList. If found then ranks the word and creates link
  @param AWord Word to parse
  @return Parsed word (or not, if not found)
-----------------------------------------------------------------------------.}
function TFileItem.ParseWord(AWord: String): String;
var
  LIsPoint: Boolean;
  LStr, LStrSave, s: String;
  LItem: TFileItem;
  Ranks: array [0..65535] of TRankItem;
  RanksCount, LMaxRank: Integer;
  i, j: Integer;
begin
  // Cut the point
  if AWord[Length(AWord)] = '.' then begin
    LStrSave := Copy(AWord, 1, Length(AWord)-1);
    LIsPoint := True;
  end else begin
    LStrSave := AWord;
    LIsPoint := False;
  end;
  Result := UpperCase(LStrSave);

  // Parse complex word (Result - item name, LStr - left rest of AWord)
  LStr := Result;
  while Pos('.', Result) > 0 do
    Delete(Result, 1, Pos('.', Result));
  Delete(LStr, Pos(Result, LStr), Length(Result));
  if Length(LStr) > 0 then Delete(LStr, Length(LStr), 1);

  // If the word is current object name then exit
  if Result = UpperName then begin
    Result := AWord;
    Exit;
  end;

  // Find first element with the name in IList
  i := FindFileItem(Result);
  if i = CNotFound then begin // If not found then exit
    Result := AWord;
    Exit;
  end;

  // Read all items with given name to Ranks list
  RanksCount := 0;
  repeat
    Inc(RanksCount);
    Ranks[RanksCount-1].Item := IList.FileItems[i];
    Ranks[RanksCount-1].Rank := 0;
    Inc(i);
  until (i >= IList.Count-1) or (IList.FileItems[i-1].UpperName <> IList.FileItems[i].UpperName);

  // Rank objects in Ranks list
  if LStr = '' then

    for i:=0 to RanksCount-1 do begin

    // Single word ranking
      // 20 points for each of common owner
      LItem := Self;
      while LItem <> nil do begin
        if Ranks[i].Item.IsMyOwner(LItem) then Inc(Ranks[i].Rank, 20);
        LItem := LItem.Owner;
      end;
      // 15 point for inheritance (one time)
      if IsMyInheritance(Ranks[i].Item) then Inc(Ranks[i].Rank, 15);

    end else

    for i:=0 to RanksCount-1 do begin

    // Complex word (LStr.Result) ranking
      // 12 points if found
      s := Ranks[i].Item.GetFullName;
      s := Copy(s, Length(s)-Length(LStr)-Length(Ranks[i].Item.Name), Length(LStr));
      if UpperCase(s) = LStr then Inc(Ranks[i].Rank, 12);

    end;

  // Debug log
  // Removed, causes VERY slow parse process, even the Options[COptDebug].Vale = False, WHY ????????
{  if Options[COptDebug].Value then begin
    Result := AWord + ', instances=' + IntToStr(RanksCount);
    for i:=0 to RanksCount-1 do
      Result := Result + ', ' + Ranks[i].Item.GetFullName + '=' + IntToStr(Ranks[i].Rank);
    AddDebug(Result);
  end; }

  // Search nearest item
  LMaxRank := 0;
  j := 0;
  for i:=0 to RanksCount-1 do // Search the largest rank
    if Ranks[i].Rank > LMaxRank then begin
      LMaxRank := Ranks[i].Rank;
      j := i;
    end;

  // If theres more than one the same rank then don't create link (exit) - only if not COptUncertain (-L option)
  if not Options[COptUncertain].Value then
    for i:=0 to RanksCount-1 do
      if (Ranks[i].Rank = LMaxRank) and (i <> j) then begin
        Result := AWord;
        Exit;
      end;

  // Create link
  Result := Format(CLinkFormat, [Ranks[j].Item.FileName + '.' + CFilenameExtension, Ranks[j].Item.GetFullName, LStrSave]);

  // Add a point (if needed)
  if LIsPoint then Result := Result + '.';
end;

{.-----------------------------------------------------------------------------
  Parses file contents
-----------------------------------------------------------------------------.}
procedure TFileItem.Parse;
var
  LEndPos: Integer;
  LTagType: TTagType;
  LParseTag: Boolean;
  s, t: String;
  i: Integer;
begin
  AddDebug('parse'); // Debug log

  LTagType := ttUnknown;

  while Length(FContents) > 0 do begin

    if FContents[1] = '<' then begin
      LEndPos := Pos('>', FContents);
      LParseTag := True;
    end else begin
      if Pos('<', FContents) > 0 then LEndPos := Pos('<', FContents) - 1 else LEndPos := Length(FContents);
      LParseTag := False;
    end;

    s := Copy(FContents, 1, LEndPos);
    Delete(FContents, 1, LEndPos);

    if LParseTag then begin // Pointer is inside HTML tag
      FParsed := FParsed + s;
      if Pos(' ', s) > 0 then
        t := UpperCase(Copy(s, 2, Pos(' ', s)-2)) else
        t := UpperCase(Copy(s, 2, Pos('>', s)-2));
      if t = 'TITLE' then LTagType := ttTitle else
      if t = 'A' then LTagType := ttA else LTagType := ttUnknown;

    end else begin // Pointer is outside HTML tag
      case LTagType of
        ttTitle: t := GetFullName;
        ttA: t := s;
      else

           // Parse text outside HTML tag
          t := '';
          LEndPos := 0;

          for i:=1 to Length(s) do begin

            if (s[i] in CSeparators) then
              if LEndPos = 0 then t := t + s[i]
                else begin
                  t := t + ParseWord(Copy(s, LEndPos, i-LEndPos)) + s[i];
                  LEndPos := 0;
                end
            else
              if LEndPos = 0 then LEndPos := i;

          end;
          if LEndPos > 0 then t := t + ParseWord(Copy(s, LEndPos, Length(s)));

      end;
      FParsed := FParsed + t;
    end;

  end;
end;

{.-----------------------------------------------------------------------------
  Reads associated file contents to FContents and searches inheritance
-----------------------------------------------------------------------------.}
procedure TFileItem.Read;
var
  f: Text;
  s: String;
  i: Integer;
begin
  AssignFile(f, SourceDir + FileName + '.' + CFilenameExtension);
  Reset(f);
  while not Eof(f) do begin
    ReadLn(f, s);
    FContents := FContents + s;
    if not Eof(f) then FContents := FContents + #13 + #10;
  end;
  CloseFile(f);

  // Search the ancestor (if available)
  if Pos(CInheritance, FContents) <> 0 then begin
    s := Copy(FContents, Pos(CInheritance, FContents)+Length(CInheritance), Length(FContents));
    s := UpperCase(Copy(s, 1, Pos('<', s)-1));

    for i:=0 to IList.Count-1 do
      if (UpperCase(IList.FileItems[i].Name) = s) and (IList.FileItems[i].OwnerCount = 2) and (IList.FileItems[i] is TFileContainer) then begin
        FInheritance := IList.FileItems[i];
        break;
      end;
  end;

  // Debug log
  if FInheritance <> nil then
    AddDebug('read, ancestor=' + FInheritance.FName) else
    AddDebug('read');
end;

{.-----------------------------------------------------------------------------
  Writes parsed file contents from FParsed to destination file
-----------------------------------------------------------------------------.}
procedure TFileItem.Write;
var
  f: Text;
begin
  AddDebug('write'); // Debug log

  AssignFile(f, DestDir + FileName + '.' + CFilenameExtension);
  Rewrite(f);
  WriteLn(f, FParsed);
  CloseFile(f);
end;

{.-----------------------------------------------------------------------------
  Returns full item name (with owner's name, with '.' as delimeter)
  @return Full item name
-----------------------------------------------------------------------------.}
function TFileItem.GetFullName: String;
var
  LOwner: TFileItem;
begin
  LOwner := Self;
  Result := '';
  while LOwner <> nil do begin
    Result := LOwner.Name + Result;
    LOwner := LOwner.FOwner;
    if LOwner <> nil then Result := '.' + Result;
  end;
end;

{.-----------------------------------------------------------------------------
  Checks the given item is one of the item owner (in files structore)
  @param AItem Item to check
  @return TRUE if AItem is one of the item owner
  @return FALSE otherwise
-----------------------------------------------------------------------------.}
function TFileItem.IsMyOwner(AItem: TFileItem): Boolean;
var
  LItem: TFileItem;
begin
  Result := False;

  if AItem = nil then Exit;
  LItem := Owner;

  while (LItem <> nil) and not Result do begin
    Result := (LItem = AItem);
    LItem := LItem.Owner;
  end;
end;

{.-----------------------------------------------------------------------------
  Checks the given item is one of the item ancestor (in classes structure)
  @param AItem Item to check
  @return TRUE if AItem is one of the item ancestor
  @return FALSE otherwise
-----------------------------------------------------------------------------.}
function TFileItem.IsMyInheritance(AItem: TFileItem): Boolean;
var
  LInheritance: TFileItem;
begin
  Result := False;

  if AItem = nil then Exit;
  if Inheritance <> nil then
    LInheritance := Inheritance else
    if (Owner <> nil) and (Owner.Inheritance <> nil) then
      LInheritance := Owner.Inheritance else
      Exit;

  while (LInheritance <> nil) and not Result do begin
    Result := (LInheritance = AItem) or ((AItem.Owner <> nil) and (LInheritance = AItem.Owner));
    LInheritance := LInheritance.Inheritance;
  end;
end;

{.-----------------------------------------------------------------------------
  Writes debug log to DebugFile
  @param DebugText Debug log
-----------------------------------------------------------------------------.}
procedure TFileItem.AddDebug(DebugText: String);
var
  i: Integer;
  s: String;
begin
  if Options[COptDebug].Value then begin // Only if +D option
    s := '  ';
    for i:=0 to OwnerCount do s := s + ' ';
    WriteLn(DebugFile, s + '[' + FFileName + '] ' + DebugText);
  end;
end;

{ TFileContainer }

{.-----------------------------------------------------------------------------
  Object contructor
  Calls inherited Create and creates all descendant items in FItemList
  @param AFilename File name to create object from
  @param AOwner Item owner (in files structure)
-----------------------------------------------------------------------------.}
constructor TFileContainer.Create(AFilename: TFilename; AOwner: TFileItem);
begin
  inherited Create(AFilename, AOwner);
  FItemList := TItemList.Create(True);

  while (ListPosition < FileList.Count-1) and (IsMySubItem(ListPosition)) do
    if HasSubItem(ListPosition) then
      FItemList.Add(TFileContainer.Create(FileList.Strings[ListPosition], Self)) else
      FItemList.Add(TFileItem.Create(FileList.Strings[ListPosition], Self));
end;

{.-----------------------------------------------------------------------------
  Object destructor
  Calls inherited Destroy and destroys FItemList
-----------------------------------------------------------------------------.}
destructor TFileContainer.Destroy;
begin
  AddDebug('destroy item list'); // Debug log
  FItemList.Free;
  inherited;
end;

{.-----------------------------------------------------------------------------
  Parses item (by inherited method) and descendant items in FItemList
-----------------------------------------------------------------------------.}
procedure TFileContainer.Parse;
var
  i: Integer;
begin
  inherited Parse;
  for i:=0 to FItemList.Count-1 do FItemList.FileItems[i].Parse;
end;

{.-----------------------------------------------------------------------------
  Reads file contents (by inherited method) and descendant items' file contents (in FItemList)
-----------------------------------------------------------------------------.}
procedure TFileContainer.Read;
var
  i: Integer;
begin
  inherited Read;
  for i:=0 to FItemList.Count-1 do FItemList.FileItems[i].read;
end;

{.-----------------------------------------------------------------------------
  Writes parsed file (by inherited method) and descendant items' parsed file (in FItemList)
-----------------------------------------------------------------------------.}
procedure TFileContainer.Write;
var
  i: Integer;
begin
  inherited Write;
  for i:=0 to FItemList.Count-1 do FItemList.FileItems[i].Write;
end;

{.-----------------------------------------------------------------------------
  Checks the given file is the item's associated file subitem
  @param Index Index of file in FileList
  @return TRUE if given file is item's associated file subitem
  @return FALSE otherwise
-----------------------------------------------------------------------------.}
function TFileContainer.IsMySubItem(Index: Integer): Boolean;
begin
  Result := False;
  if Pos(FFilename + CFilenameSeparator, FileList.Strings[Index]) <> 0 then Result := True;
end;

end.
